package ap.proof.goal;

import ap.util.Debug
import ap.basetypes.LeftistHeap

object TaskManager {
  
  private def AC = Debug.AC_GOAL
  
  private implicit def orderTask(thisTask : Task) : Ordered[Task] =
    new Ordered[Task] {
      def compare(thatTask : Task) : Int =
        (thisTask.priority compare thatTask.priority)
    }
  
  val EMPTY : TaskManager = new TaskManager (LeftistHeap.EMPTY_HEAP[Task])
  
}

/**
 * An immutable class (priority queue) for handling a set of tasks in a proof
 * goal. Currently, this is implemented using a sorted set, but it would be
 * better to use a real immutable queue (leftist heap?).
 *
 * TODO: So far, no real subsumption checks are performed
 */
class TaskManager private (tasks : LeftistHeap[Task]) extends Iterable[Task] {

  def +(t : Task) = new TaskManager (tasks + t)

  def ++ (elems: Iterable[Task]): TaskManager = this ++ elems.elements

  def ++ (elems: Iterator[Task]): TaskManager = new TaskManager (tasks ++ elems)

  def enqueue(elems: Task*): TaskManager = (this ++ elems)

  /** Returns the element with the smallest priority value in the queue,
   *  and removes this element from the queue.
   *
   *  @throws Predef.NoSuchElementException
   *  @return   the element with the smallest priority and the new task manager.
   */
  def dequeue: (Task, TaskManager) = {
    val res = tasks.findMin
    val newManager = new TaskManager(tasks.deleteMin)
    (res, newManager)
  }

  /**
   * Remove the first task from the queue, and together with it remove all
   * equal tasks that occur in the beginning of the queue
   */
  def removeFirst : TaskManager = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(TaskManager.AC, !isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    
    val head = tasks.findMin
    var res = tasks.deleteMin
    while (!res.isEmpty && res.findMin == head) res = res.deleteMin
    new TaskManager (res)
  }
  
  /** Returns the element with the smallest priority value in the queue,
   *  or throws an error if there is no element contained in the queue.
   *
   *  @return   the element with the smallest priority.
   */
  def max: Task = tasks.findMin

  //////////////////////////////////////////////////////////////////////////////
  
  def length : Int = tasks.size
  
  def elements : Iterator[Task] = tasks.elements
  
  //////////////////////////////////////////////////////////////////////////////
  
  override def toString : String = tasks.toString
  
}
