package ap.util;

import scala.collection.mutable.PriorityQueue

/**
 * Priority queue that can handle both single elements and pre-sorted sequences
 * (iterators) of elements
 */
class PriorityQueueWithIterators[A <% Ordered[A]] extends PeekIterator[A] {

  private val AC = Debug.AC_QUEUE_WITH_ITERATORS

  private implicit def orderIterator(thisIt : PeekIterator[A])
                                                   : Ordered[PeekIterator[A]] =
    new Ordered[PeekIterator[A]] {
      def compare(thatIt : PeekIterator[A]) : Int = {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(AC, thisIt.hasNext && thatIt.hasNext)
        ////////////////////////////////////////////////////////////////////////
        thisIt.peekNext compare thatIt.peekNext
      }
    }
  
  /** the internal priority queue holding the single elements */
  private val singleElements = new PriorityQueue[A]
  
  /** the internal priority queue holding sequences of elements */
  private val sequences = new PriorityQueue[PeekIterator[A]]

  //////////////////////////////////////////////////////////////////////////////
  
  private var maxElement : A = _
  private var maxElementAvailable : Boolean = false
  
  private def enqueueMaxElement : Unit = {
    if (maxElementAvailable) {
      singleElements += maxElement
      maxElementAvailable = false
    }    
  }
  
  def peekNext : A = {
    if (!maxElementAvailable) {
      def peekNextFromSequence = {
        val maxSeq = sequences.dequeue
        maxElement = maxSeq.next
        if (maxSeq.hasNext) {
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(AC, maxElement >= maxSeq.peekNext)
          //////////////////////////////////////////////////////////////////////
          sequences += maxSeq
        }
      }
      
      if (singleElements.isEmpty) {
        peekNextFromSequence // this throws an exception if the queue
                             // <code>sequences</code> is empty
      } else {
        if (sequences.isEmpty) {
          maxElement = singleElements.dequeue
        } else {
          if (sequences.max.peekNext > singleElements.max) {
            peekNextFromSequence
          } else {
            maxElement = singleElements.dequeue            
          }
        }        
      }
      
      // only reachable if an element was available (otherwise an exception was
      // thrown)
      maxElementAvailable = true
    }

    maxElement
  }
  
  def next : A = {
    val res = peekNext
    maxElementAvailable = false
    res
  }
  
  def hasNext : Boolean = 
    maxElementAvailable || !singleElements.isEmpty || !sequences.isEmpty
  
  //////////////////////////////////////////////////////////////////////////////
  
  /** Inserts a single element into the priority queue. */
  def +=(elem: A): Unit = {
    enqueueMaxElement
    singleElements += elem
  }

  /** Inserts a sequence as a single element into the priority queue. The
   * given iterator has to produce elements in descending order! */
  def +=(it: Iterator[A]): Unit = {
    if (it.hasNext) {
      enqueueMaxElement
      sequences += PeekIterator(it)
    }
  }

  /** Adds all elements provided by an iterable into the priority queue. */
  def ++=(it: Iterable[A]): Unit = (this ++= it.elements)

  /** Adds all elements provided by an iterator into the priority queue. */
  def ++=(it: Iterator[A]): Unit = it foreach { e => this += e }
  
  /** Adds all elements to the queue. */
  def enqueue(elems: A*): Unit = (this ++= elems)

  /** Adds all elements to the queue. The
   * given iterators have to produce elements in descending order! */
  def enqueue(it: Iterator[A]): Unit = (this += it)
  
  /** Returns the element with the highest priority in the queue,
   *  and removes this element from the queue.
   *
   *  @throws Predef.NoSuchElementException
   *  @return   the element with the highest priority.
   */
  def dequeue: A = next
  
  /** Returns the element with the highest priority in the queue,
   *  or throws an error if there is no element contained in the queue.
   *
   *  @return   the element with the highest priority.
   */
  def max: A = peekNext
  
  /** Removes all elements from the queue. After this operation is completed,
   *  the queue will be empty.
   */
  def clear(): Unit = {
    maxElementAvailable = false
    singleElements.clear
    sequences.clear
  }

  /** Returns whether this queue contains any elements */
  def isEmpty : Boolean = !hasNext
}
